@@ -3,22 +3,22 @@ from logging import Logger
 import time
 from os import environ
 from typing import Optional, Sequence, Dict, Union, cast, Iterator, Iterable, List
-from datetime import timedelta
+
 from api_session import APISession, JSONDict
 import requests
 from requests.exceptions import HTTPError
 
 from magento.types import Product, SourceItem, Sku, Category, MediaEntry, MagentoEntity, Order, PathId
 from magento.version import __version__
-from magento.exceptions import MagentoException
+from magento.exceptions import MagentoException, MagentoAssertionError
 
 from magento.queries import Query, make_search_query, make_field_value_query
 
 __all__ = (
-    'Magento',
+    "Magento",
 )
 
-USER_AGENT = f'Bixoto/PyMagento {__version__} +git.io/JDp0h'
+USER_AGENT = f"Bixoto/PyMagento {__version__} +git.io/JDp0h"
 
 DEFAULT_ATTRIBUTE_DICT = {
     "apply_to": [],
@@ -63,7 +63,7 @@ def raise_for_response(response: requests.Response):
     if response.ok:
         return
 
-    if response.text and response.text[0] == '{':
+    if response.text and response.text[0] == "{":
         try:
             body = response.json()
         except (ValueError, JSONDecodeError):
@@ -76,19 +76,22 @@ def raise_for_response(response: requests.Response):
     response.raise_for_status()
 
 
+def escape_path(sku: str):
+    return sku.replace("%", "%25").replace("/", "%2F")
+
+
 class Magento(APISession):
     """
     Client for the Magento API.
     """
     # default batch size for paginated requests
     # Note increasing it doesn’t create a significant time improvement.
-    # For example, in one test on Bixoto production in 2021, getting 2k products using a page size of 1k took 28s.
+    # For example, in one test on Bixoto in 2021, getting 2k products using a page size of 1k took 28s.
     # The same query with a page size of 2k still took 26s.
+    # Magento supports setting hard limits on this:
+    #   https://developer.adobe.com/commerce/webapi/get-started/api-security/
     PAGE_SIZE = 1000
 
-    # default is 4 hours for admin tokens (the ones we use)
-    TOKEN_LIFETIME = timedelta(hours=3)
-
     def __init__(self,
                  token: Optional[str] = None,
                  base_url: Optional[str] = None,
@@ -96,9 +99,9 @@ class Magento(APISession):
                  logger: Optional[Logger] = None,
                  read_only=False,
                  user_agent=None,
-                 verbose=False,
-                 # Deprecated
-                 log_progress=False):
+                 *,
+                 batch_page_size: Optional[int] = None,
+                 **kwargs):
         """
         Create a Magento client instance. All arguments are optional and fall back on environment variables named
         ``PYMAGENTO_ + argument.upper()`` (``PYMAGENTO_TOKEN``, ``PYMAGENTO_BASE_URL``, etc).
@@ -106,12 +109,11 @@ class Magento(APISession):
 
         :param token: API integration token
         :param base_url: base URL of the Magento instance
-        :param scope: API scope
+        :param scope: API scope. Default on ``PYMAGENTO_SCOPE`` if set, or ``"all"``
+        :param batch_page_size: if set, override the default page size used for batch queries.
         :param logger: optional logger.
-        :param verbose: if True, log the progress of get_paginated. This has no effect if logger is not set.
-        :param read_only: if True,
+        :param read_only:
         :param user_agent: User-Agent
-        :param log_progress: deprecated alias for `verbose`.
         """
         token = token or environ.get("PYMAGENTO_TOKEN")
         base_url = base_url or environ.get("PYMAGENTO_BASE_URL")
@@ -123,15 +125,14 @@ class Magento(APISession):
         if base_url is None:
             raise RuntimeError("Missing API base URL")
 
-        super().__init__(base_url=base_url, user_agent=user_agent, read_only=read_only)
+        super().__init__(base_url=base_url, user_agent=user_agent, read_only=read_only, **kwargs)
 
-        if log_progress:
-            verbose = True
+        if batch_page_size is not None:
+            self.PAGE_SIZE = batch_page_size
 
         self.scope = scope
         self.logger = logger
-        self.verbose = verbose if logger else False
-        self.headers['Authorization'] = f"Bearer {token}"
+        self.headers["Authorization"] = f"Bearer {token}"
 
     # Attributes
     # ==========
@@ -142,17 +143,17 @@ class Magento(APISession):
             base.update(attribute)
             attribute = base
 
-        return self.post_api('/V1/products/attributes', json={"attribute": attribute}, throw=throw, **kwargs).json()
+        return self.post_api("/V1/products/attributes", json={"attribute": attribute}, throw=throw, **kwargs).json()
 
     def delete_attribute(self, attribute_code: str, **kwargs):
-        return self.delete_api(f"/V1/products/attributes/{attribute_code}", **kwargs)
+        return self.delete_api(f"/V1/products/attributes/{escape_path(attribute_code)}", **kwargs)
 
     # Attribute Sets
     # ==============
 
-    def get_attribute_sets(self, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
+    def get_attribute_sets(self, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
         """Get all attribute sets (generator)."""
-        return self.get_paginated("/V1/eav/attribute-sets/list", limit=limit, **kwargs)
+        return self.get_paginated("/V1/eav/attribute-sets/list", query=query, limit=limit, **kwargs)
 
     def get_attribute_set_attributes(self, attribute_set_id: int, **kwargs):
         """Get all attributes for the given attribute set id."""
@@ -179,7 +180,8 @@ class Magento(APISession):
         return self.post_api("/V1/products/attribute-sets/attributes", json=payload, **kwargs)
 
     def remove_attribute_set_attribute(self, attribute_set_id: int, attribute_code: str, **kwargs):
-        return self.delete_api(f"/V1/products/attribute-sets/{attribute_set_id}/attributes/{attribute_code}", **kwargs)
+        path = f"/V1/products/attribute-sets/{attribute_set_id}/attributes/{escape_path(attribute_code)}"
+        return self.delete_api(path, **kwargs)
 
     # Bulk Operations
     # ===============
@@ -188,33 +190,27 @@ class Magento(APISession):
         """
         Get the status of an async/bulk operation.
         """
-        return self.get_api(f'/V1/bulk/{bulk_uuid}/status', throw=True).json()
+        return self.get_api(f"/V1/bulk/{escape_path(bulk_uuid)}/status", throw=True).json()
 
     # Carts
     # =====
 
-    def get_carts(self, *, query: Query = None, limit=-1) -> Iterable[MagentoEntity]:
+    def get_carts(self, *, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
         """Get all carts (generator)."""
-        return self.get_paginated("/V1/carts/search", query=query, limit=limit)
+        return self.get_paginated("/V1/carts/search", query=query, limit=limit, **kwargs)
 
     # Categories
     # ==========
 
-    def get_categories(self, query: Query = None) -> Iterable[Category]:
+    def get_categories(self, query: Query = None, limit=-1, **kwargs) -> Iterable[Category]:
         """
         Yield all categories.
-
-        :param query:
-        :return:
         """
-        return self.get_paginated('/V1/categories/list', query=query)
+        return self.get_paginated("/V1/categories/list", query=query, limit=limit, **kwargs)
 
     def get_category(self, category_id: PathId) -> Optional[Category]:
         """
         Return a category given its id.
-
-        :param category_id:
-        :return:
         """
         return self.get_json_api(f"/V1/categories/{category_id}")
 
@@ -238,47 +234,47 @@ class Magento(APISession):
         :param category_data: (partial) category data to update
         :return: updated category
         """
-        return cast(Category, self.put_api(f'/V1/categories/{category_id}',
+        return cast(Category, self.put_api(f"/V1/categories/{category_id}",
                                            json={"category": category_data}, throw=True).json())
 
     def create_category(self, category: Category, throw=False):
         """
         Create a new category.
         """
-        return self.post_api('/V1/categories', json={"category": category}, throw=throw)
+        return self.post_api("/V1/categories", json={"category": category}, throw=throw)
 
     # CMS
     # ===
 
-    def get_cms_pages(self, *, query: Query = None, limit=-1) -> Iterable[MagentoEntity]:
+    def get_cms_pages(self, *, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
         """Get all CMS pages (generator)."""
-        return self.get_paginated("/V1/cmsPage/search", query=query, limit=limit)
+        return self.get_paginated("/V1/cmsPage/search", query=query, limit=limit, **kwargs)
 
-    def get_cms_blocks(self, *, query: Query = None, limit=-1) -> Iterable[MagentoEntity]:
+    def get_cms_blocks(self, *, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
         """Get all CMS blocks (generator)."""
-        return self.get_paginated("/V1/cmsBlock/search", query=query, limit=limit)
+        return self.get_paginated("/V1/cmsBlock/search", query=query, limit=limit, **kwargs)
 
     # Coupons
     # =======
 
-    def get_coupons(self, *, query: Query = None, limit=-1) -> Iterable[MagentoEntity]:
+    def get_coupons(self, *, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
         """Get all coupons (generator)."""
-        return self.get_paginated("/V1/coupons/search", query=query, limit=limit)
+        return self.get_paginated("/V1/coupons/search", query=query, limit=limit, **kwargs)
 
     # Customers
     # =========
 
-    def get_customers(self, *, query: Query = None, limit=-1) -> Iterable[MagentoEntity]:
+    def get_customers(self, *, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
         """Get all customers (generator)."""
-        return self.get_paginated("/V1/customers/search", query=query, limit=limit)
+        return self.get_paginated("/V1/customers/search", query=query, limit=limit, **kwargs)
 
     def get_customer(self, customer_id: int) -> dict:
         """Return a single customer."""
         return self.get_api(f"/V1/customers/{customer_id}", throw=True).json()
 
-    def get_customer_groups(self, *, query: Query = None, limit=-1) -> Iterable[MagentoEntity]:
+    def get_customer_groups(self, *, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
         """Get all customer groups (generator)."""
-        return self.get_paginated("/V1/customerGroups/search", query=query, limit=limit)
+        return self.get_paginated("/V1/customerGroups/search", query=query, limit=limit, **kwargs)
 
     # Invoices
     # ========
@@ -312,9 +308,9 @@ class Magento(APISession):
             return invoice
         return None
 
-    def get_invoices(self, query: Query = None, limit=-1) -> Iterable[MagentoEntity]:
+    def get_invoices(self, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
         """Get all invoices (generator)."""
-        return self.get_paginated("/V1/invoices", query=query, limit=limit)
+        return self.get_paginated("/V1/invoices", query=query, limit=limit, **kwargs)
 
     def get_order_invoices(self, order_id: Union[int, str]):
         """Get invoices for the given order id."""
@@ -343,38 +339,36 @@ class Magento(APISession):
         if status:
             query = make_field_value_query("status", status, condition_type=status_condition_type)
 
-        return self.get_paginated("/V1/orders", limit=limit, query=query, retry=retry)
+        return self.get_paginated("/V1/orders", query=query, limit=limit, retry=retry)
 
     def get_last_orders(self, limit=10) -> List[Order]:
         """Return a list of the last orders (default: 10)."""
         query = make_search_query([], sort_orders=[("increment_id", "DESC")])
         return list(self.get_orders(query=query, limit=limit))
 
-    def get_orders_items(self, *, sku: Optional[str] = None, query: Query = None, **kwargs):
+    def get_orders_items(self, *, sku: Optional[str] = None, query: Query = None, limit=-1, **kwargs):
         """
         Return orders items.
 
         :param sku: filter orders items on SKU. This is a shortcut for ``query=make_field_value_query("sku", sku)``.
         :param query: optional query. This take precedence over ``sku``.
+        :param limit:
         :return:
         """
         if query is None and sku is not None:
             query = make_field_value_query("sku", sku)
 
-        return self.get_paginated("/V1/orders/items", query=query, **kwargs)
+        return self.get_paginated("/V1/orders/items", query=query, limit=limit, **kwargs)
 
     def get_order(self, order_id: str, throw=True) -> Optional[Order]:
         """
         Get an order given its (entity) id.
         """
-        return self.get_api(f'/V1/orders/{order_id}', throw=throw).json()
+        return self.get_api(f"/V1/orders/{order_id}", throw=throw).json()
 
     def get_order_by_increment_id(self, increment_id: str) -> Optional[Order]:
         """
         Get an order given its increment id. Return ``None`` if the order doesn’t exist.
-
-        :param increment_id:
-        :return:
         """
         query = make_field_value_query("increment_id", increment_id)
         for order in self.get_orders(query=query, limit=1):
@@ -382,15 +376,35 @@ class Magento(APISession):
         return None
 
     def hold_order(self, order_id: str, **kwargs):
-        return self.post_api(f'/V1/orders/{order_id}/hold', **kwargs)
+        """
+        Hold an order. This is the opposite of ``unhold_order``.
+
+        :param order_id: order id (not increment id)
+        """
+        return self.post_api(f"/V1/orders/{order_id}/hold", **kwargs)
 
     def unhold_order(self, order_id: str, **kwargs):
-        return self.post_api(f'/V1/orders/{order_id}/unhold', **kwargs)
+        """
+        Un-hold an order. This is the opposite of ``hold_order``.
+
+        :param order_id: order id (not increment id)
+        """
+        return self.post_api(f"/V1/orders/{order_id}/unhold", **kwargs)
 
     def save_order(self, order: Order):
-        return self.post_api(f'/V1/orders', json={"entity": order})
+        """Save an order."""
+        return self.post_api(f"/V1/orders", json={"entity": order})
 
     def set_order_status(self, order: Order, status: str, *, external_order_id: Optional[str] = None):
+        """
+        Change the status of an order, and optionally set its ``ext_order_id``. This is a convenient wrapper around
+        ``save_order``.
+
+        :param order: order payload
+        :param status: new status
+        :param external_order_id: optional external order id
+        :return:
+        """
         payload = {
             "entity_id": order["entity_id"],
             "status": status,
@@ -404,9 +418,9 @@ class Magento(APISession):
     # Credit Memos
     # ============
 
-    def get_credit_memos(self, query: Query = None, limit=-1) -> Iterable[MagentoEntity]:
+    def get_credit_memos(self, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
         """Get all credit memos (generator)."""
-        return self.get_paginated("/V1/creditmemos", query=query, limit=limit)
+        return self.get_paginated("/V1/creditmemos", query=query, limit=limit, **kwargs)
 
     # Prices
     # ======
@@ -440,11 +454,8 @@ class Magento(APISession):
     def get_special_prices(self, skus: Sequence[Sku]) -> List[MagentoEntity]:
         """
         Get special prices for a sequence of SKUs.
-
-        :param skus:
-        :return:
         """
-        return self.post_api('/V1/products/special-price-information',
+        return self.post_api("/V1/products/special-price-information",
                              json={"skus": skus}, throw=True, bypass_read_only=True).json()
 
     def save_special_prices(self, special_prices: Sequence[MagentoEntity]):
@@ -461,23 +472,17 @@ class Magento(APISession):
         :param special_prices: Special prices to save.
         :return:
         """
-        return self.post_api('/V1/products/special-price', json={"prices": special_prices})
+        return self.post_api("/V1/products/special-price", json={"prices": special_prices})
 
     def delete_special_prices(self, special_prices: Sequence[MagentoEntity]):
         """
         Delete a sequence of special prices.
-
-        :param special_prices:
-        :return:
         """
-        return self.post_api('/V1/products/special-price-delete', json={"prices": special_prices})
+        return self.post_api("/V1/products/special-price-delete", json={"prices": special_prices})
 
     def delete_special_prices_by_sku(self, skus: Sequence[Sku]):
         """
-        Equivalent of `delete_special_prices(get_special_prices(skus))`.
-
-        :param skus:
-        :return:
+        Equivalent of ``delete_special_prices(get_special_prices(skus))``.
         """
         special_prices = self.get_special_prices(skus)
         return self.delete_special_prices(special_prices)
@@ -494,26 +499,26 @@ class Magento(APISession):
         :param retry:
         :return:
         """
-        return cast(Iterator[Product], self.get_paginated("/V1/products/", limit=limit, query=query, retry=retry))
+        return cast(Iterator[Product], self.get_paginated("/V1/products/", query=query, limit=limit, retry=retry))
 
     def get_products_types(self) -> Sequence[MagentoEntity]:
         """Get available product types."""
-        return self.get_json_api('/V1/product/types')
+        return self.get_json_api("/V1/product/types")
 
     def get_product(self, sku: Sku) -> Optional[Product]:
         """
         Get a single product. Return ``None`` if it doesn’t exist.
 
-        :param sku:
+        :param sku: SKU of the product
         :return:
         """
-        return self.get_json_api(f'/V1/products/{sku}')
+        return self.get_json_api(f"/V1/products/{escape_path(sku)}")
 
     def get_product_by_id(self, product_id: int) -> Optional[Product]:
         """
         Get a product given its id. Return ``None`` if the product doesn’t exist.
 
-        :param product_id:
+        :param product_id: ID of the product
         :return:
         """
         query = make_field_value_query("entity_id", product_id)
@@ -521,6 +526,27 @@ class Magento(APISession):
             return product
         return None
 
+    def get_product_by_query(self, query: Query, *, expect_one=True) -> Optional[Product]:
+        """
+        Get a product with a custom query. Return ``None`` if the query doesn’t return match any product, and raise
+        an exception if it returns more than one, unless ``expect_one`` is set to ``False``.
+
+        :param query:
+        :param expect_one: if True (the default), raise an exception if the query returns more than one result.
+        :return:
+        """
+        if not expect_one:
+            for product in self.get_products(query=query, limit=1):
+                return product
+            return None
+
+        products = list(self.get_products(query=query, limit=2))
+        if not products:
+            return None
+        if len(products) == 1:
+            return products[0]
+        raise MagentoAssertionError("Got more than one product for query %r" % query)
+
     def get_product_medias(self, sku: Sku) -> Sequence[MediaEntry]:
         """
         Get the list of gallery entries associated with the given product.
@@ -528,7 +554,7 @@ class Magento(APISession):
         :param sku: SKU of the product.
         :return:
         """
-        return self.get_json_api(f'/V1/products/{sku}/media')
+        return self.get_json_api(f"/V1/products/{escape_path(sku)}/media")
 
     def get_product_media(self, sku: Sku, entry_id: PathId) -> MediaEntry:
         """
@@ -538,20 +564,31 @@ class Magento(APISession):
         :param entry_id:
         :return:
         """
-        return self.get_json_api(f'/V1/products/{sku}/media/{entry_id}')
+        # TODO: s/entry_id/media_id
+        return self.get_json_api(f"/V1/products/{escape_path(sku)}/media/{entry_id}")
 
     def save_product_media(self, sku: Sku, media_entry: MediaEntry):
-        return self.post_api(f'/V1/products/{sku}/media', json={"entry": media_entry}, throw=True).json()
+        """
+        Save a product media.
+        """
+        return self.post_api(f"/V1/products/{escape_path(sku)}/media", json={"entry": media_entry}, throw=True).json()
 
     def delete_product_media(self, sku: Sku, media_id: PathId, throw=False):
-        return self.delete_api(f'/V1/products/{sku}/media/{media_id}', throw=throw)
+        """
+        Delete a media associated with a product.
 
-    def save_product(self, product, *, log_response=False, save_options: Optional[bool] = None) -> Product:
+        :param sku: SKU of the product
+        :param media_id:
+        :param throw:
+        :return:
+        """
+        return self.delete_api(f"/V1/products/{escape_path(sku)}/media/{media_id}", throw=throw)
+
+    def save_product(self, product, *, save_options: Optional[bool] = None) -> Product:
         """
         Save a product.
 
-        :param product: (partial) product to save.
-        :param log_response: if True and ``self.logger`` is set, log the response.
+        :param product: product to save (can be partial).
         :param save_options: set the `saveOptions` attribute.
         :return:
         """
@@ -560,9 +597,9 @@ class Magento(APISession):
             payload["saveOptions"] = save_options
 
         # throw=False so the log is printed before we raise
-        resp = self.post_api('/V1/products', json=payload, throw=False)
-        if log_response and self.logger:
-            self.logger.info("Save product response: %s", resp.text)
+        resp = self.post_api("/V1/products", json=payload, throw=False)
+        if self.logger:
+            self.logger.debug("Save product response: %s", resp.text)
         raise_for_response(resp)
         return cast(Product, resp.json())
 
@@ -586,20 +623,20 @@ class Magento(APISession):
         if save_options is not None:
             payload["saveOptions"] = save_options
 
-        return cast(Product, self.put_api(f'/V1/products/{sku}', json=payload, throw=True).json())
+        return cast(Product, self.put_api(f"/V1/products/{escape_path(sku)}", json=payload, throw=True).json())
 
     def delete_product(self, sku: Sku, skip_missing=False, throw=True, **kwargs) -> bool:
         """
         Delete a product given its SKU.
 
         :param sku:
-        :param skip_missing: if true, don't raise if the product is missing, and return False.
+        :param skip_missing: if true, don’t raise if the product is missing, and return False.
         :param throw: throw on error response
         :param kwargs: keyword arguments passed to all underlying methods.
         :return: a boolean indicating success.
         """
         try:
-            response = self.delete_api(f'/V1/products/{sku}', throw=throw, **kwargs)
+            response = self.delete_api(f"/V1/products/{escape_path(sku)}", throw=throw, **kwargs)
         except (HTTPError, MagentoException) as e:
             if skip_missing and e.response is not None and e.response.status_code == 404:
                 return False
@@ -618,17 +655,11 @@ class Magento(APISession):
 
         See https://devdocs.magento.com/guides/v2.4/rest/bulk-endpoints.html
 
-        :param product_updates: sequence of product data dicts. They MUST contain an 'sku' key.
+        :param product_updates: sequence of product data dicts. They MUST contain an `sku` key.
         :return:
         """
         payload = [{"product": product_update} for product_update in product_updates]
-        return self.put_api('/V1/products/bySku', json=payload, throw=True, async_bulk=True).json()
-
-    def get_product_source_items(self, sku) -> Iterable[SourceItem]:
-        """
-        Deprecated. Use `.get_source_items(sku=sku)` instead, which is equivalent.
-        """
-        return self.get_source_items(sku=sku)
+        return self.put_api("/V1/products/bySku", json=payload, throw=True, async_bulk=True).json()
 
     def set_product_stock_item(self, sku: Sku, quantity: int, is_in_stock=1):
         """
@@ -638,15 +669,15 @@ class Magento(APISession):
         :return: requests.Response
         """
         payload = {"stockItem": {"qty": quantity, "is_in_stock": is_in_stock}}
-        return self.put_api(f'/V1/products/{sku}/stockItems/1', json=payload, throw=True)
+        return self.put_api(f"/V1/products/{escape_path(sku)}/stockItems/1", json=payload, throw=True)
 
     def get_product_stock_status(self, sku: Sku) -> MagentoEntity:
         """Get stock status for an SKU."""
-        return self.get_api(f"/V1/stockStatuses/{sku}", throw=True).json()
+        return self.get_api(f"/V1/stockStatuses/{escape_path(sku)}", throw=True).json()
 
     def get_product_stock_item(self, sku: Sku) -> MagentoEntity:
         """Get the stock item for an SKU."""
-        return self.get_api(f"/V1/stockItems/{sku}", throw=True).json()
+        return self.get_api(f"/V1/stockItems/{escape_path(sku)}", throw=True).json()
 
     def link_child_product(self, parent_sku: Sku, child_sku: Sku, **kwargs) -> requests.Response:
         """
@@ -656,7 +687,7 @@ class Magento(APISession):
         :param child_sku: SKU of the child product
         :return: `requests.Response` object
         """
-        return self.post_api(f'/V1/configurable-products/{parent_sku}/child',
+        return self.post_api(f"/V1/configurable-products/{escape_path(parent_sku)}/child",
                              json={"childSku": child_sku}, **kwargs)
 
     def unlink_child_product(self, parent_sku: Sku, child_sku: Sku, **kwargs) -> requests.Response:
@@ -667,7 +698,8 @@ class Magento(APISession):
         :param child_sku: SKU of the child product
         :return: `requests.Response` object
         """
-        return self.delete_api(f"/V1/configurable-products/{parent_sku}/children/{child_sku}", **kwargs)
+        return self.delete_api(f"/V1/configurable-products/{escape_path(parent_sku)}/children/{escape_path(child_sku)}",
+                               **kwargs)
 
     def save_configurable_product_option(self, sku: Sku, option: MagentoEntity, throw=False):
         """
@@ -678,7 +710,8 @@ class Magento(APISession):
         :param throw:
         :return: `requests.Response` object
         """
-        return self.post_api(f'/V1/configurable-products/{sku}/options', json={"option": option}, throw=throw)
+        return self.post_api(f"/V1/configurable-products/{escape_path(sku)}/options",
+                             json={"option": option}, throw=throw)
 
     # Products Attribute Options
     # --------------------------
@@ -690,21 +723,22 @@ class Magento(APISession):
         :param attribute_code:
         :return: sequence of option dicts.
         """
-        response = self.get_api(f'/V1/products/attributes/{attribute_code}/options', throw=True)
+        response = self.get_api(f"/V1/products/attributes/{escape_path(attribute_code)}/options", throw=True)
         return cast(Sequence[Dict[str, str]], response.json())
 
     def add_products_attribute_option(self, attribute_code: str, option: Dict[str, str]) -> str:
         """
         Add an option to a products attribute.
 
+        https://magento.redoc.ly/2.3.6-admin/#operation/catalogProductAttributeOptionManagementV1AddPost
+
         :param attribute_code:
-        :param option: dict with label/value keys (mandatory).
-          See https://magento.redoc.ly/2.3.6-admin/#operation/catalogProductAttributeOptionManagementV1AddPost
-          for the optional keys.
+        :param option: dict with label/value keys (mandatory)
         :return: new id
         """
         payload = {"option": option}
-        response = self.post_api(f'/V1/products/attributes/{attribute_code}/options', json=payload, throw=True)
+        response = self.post_api(f"/V1/products/attributes/{escape_path(attribute_code)}/options",
+                                 json=payload, throw=True)
         ret = cast(str, response.json())
 
         if ret.startswith("id_"):
@@ -720,7 +754,8 @@ class Magento(APISession):
         :param option_id:
         :return: boolean
         """
-        response = self.delete_api(f'/V1/products/attributes/{attribute_code}/options/{option_id}', throw=True)
+        response = self.delete_api(f"/V1/products/attributes/{escape_path(attribute_code)}/options/{option_id}",
+                                   throw=True)
         return cast(bool, response.json())
 
     # Aliases
@@ -735,22 +770,22 @@ class Magento(APISession):
     # Sales Rules
     # ===========
 
-    def get_sales_rules(self, *, query: Query = None, limit=-1) -> Iterable[MagentoEntity]:
+    def get_sales_rules(self, *, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
         """Get all sales rules (generator)."""
-        return self.get_paginated("/V1/salesRules/search", query=query, limit=limit)
+        return self.get_paginated("/V1/salesRules/search", query=query, limit=limit, **kwargs)
 
     # Shipments
     # =========
 
-    def get_shipments(self, **kwargs) -> Iterable[MagentoEntity]:
-        """Return shipments."""
-        return self.get_paginated("/V1/shipments", **kwargs)
+    def get_shipments(self, *, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
+        """Return shipments (generator)."""
+        return self.get_paginated("/V1/shipments", query=query, limit=limit, **kwargs)
 
     def ship_order(self, order_id: PathId, payload: MagentoEntity):
         """
         Ship an order.
         """
-        return self.post_api(f'/V1/order/{order_id}/ship', json=payload)
+        return self.post_api(f"/V1/order/{order_id}/ship", json=payload)
 
     def get_order_shipments(self, order_id: Union[int, str]):
         """Get shipments for the given order id."""
@@ -759,8 +794,35 @@ class Magento(APISession):
     # Stock
     # =====
 
-    def get_stock_source_links(self, query: Query = None) -> Iterable[MagentoEntity]:
-        return self.get_paginated("/V1/inventory/stock-source-links", query=query)
+    def get_stock_source_links(self, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
+        return self.get_paginated("/V1/inventory/stock-source-links", query=query, limit=limit, **kwargs)
+
+    # Sources
+    # =======
+
+    def get_sources(self, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
+        """
+        Get all sources.
+
+        https://adobe-commerce.redoc.ly/2.4.6-admin/tag/inventorysources#operation/GetV1InventorySources
+        """
+        return self.get_paginated("/V1/inventory/sources", query=query, limit=limit, **kwargs)
+
+    def get_source(self, source_code: str) -> Optional[MagentoEntity]:
+        """
+        Get a single source, or `None` if it doesn’t exist.
+
+        https://adobe-commerce.redoc.ly/2.4.6-admin/tag/inventorysourcessourceCode#operation/GetV1InventorySourcesSourceCode
+        """
+        return self.get_json_api(f"/V1/inventory/sources/{escape_path(source_code)}")
+
+    def save_source(self, source: MagentoEntity):
+        """
+        Save a source.
+
+        https://adobe-commerce.redoc.ly/2.4.6-admin/tag/inventorysources/#operation/PostV1InventorySources
+        """
+        return self.post_api("/V1/inventory/sources", json={"source": source}, throw=True).json()
 
     # Source Items
     # ============
@@ -768,7 +830,8 @@ class Magento(APISession):
     def get_source_items(self, source_code: Optional[str] = None, sku: Optional[str] = None,
                          *,
                          skus: Optional[Iterable[str]] = None,
-                         query: Query = None, limit=-1) -> Iterable[MagentoEntity]:
+                         query: Query = None, limit=-1,
+                         **kwargs) -> Iterable[MagentoEntity]:
         """
         Return a generator of all source items.
 
@@ -790,7 +853,7 @@ class Magento(APISession):
 
             query = make_search_query(filter_groups)
 
-        return self.get_paginated("/V1/inventory/source-items", limit=limit, query=query)
+        return self.get_paginated("/V1/inventory/source-items", query=query, limit=limit, **kwargs)
 
     def save_source_items(self, source_items: Sequence[SourceItem]):
         """
@@ -800,8 +863,8 @@ class Magento(APISession):
         :return:
         """
         if not source_items:
-            return
-        return self.post_api('/V1/inventory/source-items', json={"sourceItems": source_items}, throw=True).json()
+            return None
+        return self.post_api("/V1/inventory/source-items", json={"sourceItems": source_items}, throw=True).json()
 
     def delete_source_items(self, source_items: Iterable[SourceItem], throw=True, **kwargs):
         """
@@ -816,7 +879,7 @@ class Magento(APISession):
         payload = {
             "sourceItems": [{"sku": s["sku"], "source_code": s["source_code"]} for s in source_items],
         }
-        return self.post_api('/V1/inventory/source-items-delete', json=payload, throw=throw, **kwargs)
+        return self.post_api("/V1/inventory/source-items-delete", json=payload, throw=throw, **kwargs)
 
     def delete_default_source_items(self):
         """
@@ -834,24 +897,24 @@ class Magento(APISession):
     # Taxes
     # =====
 
-    def get_tax_classes(self, *, query: Query = None, limit=-1) -> Iterable[MagentoEntity]:
+    def get_tax_classes(self, *, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
         """Get all tax classes (generator)."""
-        return self.get_paginated("/V1/taxClasses/search", query=query, limit=limit)
+        return self.get_paginated("/V1/taxClasses/search", query=query, limit=limit, **kwargs)
 
-    def get_tax_rates(self, *, query: Query = None, limit=-1) -> Iterable[MagentoEntity]:
+    def get_tax_rates(self, *, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
         """Get all tax rates (generator)."""
-        return self.get_paginated("/V1/taxRates/search", query=query, limit=limit)
+        return self.get_paginated("/V1/taxRates/search", query=query, limit=limit, **kwargs)
 
-    def get_tax_rules(self, *, query: Query = None, limit=-1) -> Iterable[MagentoEntity]:
+    def get_tax_rules(self, *, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
         """Get all tax rules (generator)."""
-        return self.get_paginated("/V1/taxRules/search", query=query, limit=limit)
+        return self.get_paginated("/V1/taxRules/search", query=query, limit=limit, **kwargs)
 
     # Modules
     # =======
 
-    def get_modules(self, query: Query = None, limit=-1) -> Iterable[MagentoEntity]:
+    def get_modules(self, query: Query = None, limit=-1, **kwargs) -> Iterable[MagentoEntity]:
         """Get all enabled modules (generator)."""
-        return self.get_paginated("/V1/modules", query=query, limit=limit)
+        return self.get_paginated("/V1/modules", query=query, limit=limit, **kwargs)
 
     # Internals
     # =========
@@ -876,7 +939,7 @@ class Magento(APISession):
         full_path = f"/rest/{self.scope}"
 
         if async_bulk:
-            full_path += '/async/bulk'
+            full_path += "/async/bulk"
 
         full_path += path
 
@@ -926,15 +989,15 @@ class Magento(APISession):
             page_query["searchCriteria[currentPage]"] = current_page
 
             res = self.get_api(path, page_query, throw=True, retry=retry).json()
-            items = res.get('items', [])
+            items = res.get("items", [])
             if not items:
                 break
 
             total_count = res["total_count"]
 
             for item in items:
-                if self.verbose and self.logger and count and count % 1000 == 0:
-                    self.logger.info(f'loaded {count} items')
+                if self.logger and count and count % 1000 == 0:
+                    self.logger.debug(f"loaded {count} items")
                 yield item
                 count += 1
                 if count >= total_count:
