@@ -4,7 +4,7 @@ from typing import List, Optional, Union, Dict, cast
 
 RE_PARAM = re.compile(r"%([A-Za-z0-9]+)")
 
-__all__ = ('build_exception_text', 'MagentoException')
+__all__ = ('build_exception_text', 'MagentoException', 'MagentoAssertionError')
 
 
 def build_exception_text(message: str, parameters: Optional[Union[Dict[str, str], List[str]]]):
@@ -13,7 +13,7 @@ def build_exception_text(message: str, parameters: Optional[Union[Dict[str, str]
 
     # Magento sometimes return a dict, sometimes a list.
     # For example:
-    #   message = 'Se requiere "%fieldName".'
+    #   message = "%fieldName is a required field."
     #   parameters = {"fieldName":"product"}
     if isinstance(parameters, dict):
         return cast(str, RE_PARAM.sub(lambda m: str(parameters[m.group(1)]), message))  # type: ignore
@@ -40,3 +40,12 @@ class MagentoException(Exception):
 
         text = build_exception_text(message, parameters)
         super().__init__(text)
+
+
+class MagentoAssertionError(AssertionError):
+    """
+    Exception raised by ``Magento#get_product_by_query`` when the query returns more than one product.
+
+    This exception doesn’t inherit from ``MagentoException`` because it’s not a Magento error per se.
+    """
+    pass
